import unittest
import secrets
from pathlib import Path


class TestEntryPoint(unittest.TestCase):
    def test_plugin_available(self):
        from qiskit.transpiler.preset_passmanagers.plugin import list_stage_plugins

        scheduling_plugins = list_stage_plugins("scheduling")
        self.assertTrue("leaky_rotations" in scheduling_plugins)


class TestExample(unittest.TestCase):
    def test_example_readme_image(self):
        from qiskit.circuit.random import random_circuit
        from qiskit.transpiler.preset_passmanagers import generate_preset_pass_manager
        from qiskit.transpiler.preset_passmanagers.plugin import list_stage_plugins
        from qiskit_ibm_runtime.fake_provider import FakeKyoto
        from qiskit_leaky_scheduling import recover_data

        with open(Path(__file__).parent / "../HSLU_Logo_small.png", "rb") as file:
            hslu_logo = file.read()

        backend = FakeKyoto()
        pm = generate_preset_pass_manager(
            backend=backend,
            optimization_level=3,
            scheduling_method="leaky_rotations",
            seed_transpiler=0,
        )

        qc = random_circuit(
            num_qubits=7, depth=3, max_operands=2, measure=True, reset=False, seed=0
        )

        isa_qc = pm.run(qc)

        recovered_img = recover_data(isa_qc)[:328]
        self.assertEqual(hslu_logo, recovered_img)

    def test_example_readme_message(self):
        import builtins
        from qiskit.circuit.random import random_circuit
        from qiskit.transpiler.preset_passmanagers import generate_preset_pass_manager
        from qiskit.transpiler.preset_passmanagers.plugin import list_stage_plugins
        from qiskit_ibm_runtime.fake_provider import FakeKyoto
        from qiskit_leaky_scheduling import recover_data

        backend = FakeKyoto()
        pm = generate_preset_pass_manager(
            backend=backend,
            optimization_level=3,
            scheduling_method="leaky_rotations",
            seed_transpiler=0,
        )

        qc = random_circuit(
            num_qubits=7, depth=3, max_operands=2, measure=True, reset=False, seed=0
        )

        message = b"My secret data encoded in RZ gates."
        builtins.data = message

        isa_qc = pm.run(qc)

        recovered_data = recover_data(isa_qc)[:35]
        self.assertEqual(message, recovered_data)
        del builtins.data

    def test_random_encoding_decoding_hslu_logo(self):
        from qiskit.circuit.random import random_circuit
        from qiskit.transpiler.preset_passmanagers import generate_preset_pass_manager
        from qiskit.transpiler.preset_passmanagers.plugin import list_stage_plugins
        from qiskit_ibm_runtime.fake_provider import FakeKyiv
        from qiskit_leaky_scheduling import recover_data

        with open(Path(__file__).parent / "../HSLU_Logo_small.png", "rb") as file:
            hslu_logo = file.read()

        backend = FakeKyiv()

        for i in range(5):
            seed = secrets.randbelow(10**8)
            with self.subTest(seed=seed):
                qc = random_circuit(
                    num_qubits=50,
                    depth=5,
                    max_operands=2,
                    measure=True,
                    reset=False,
                )
                pm = generate_preset_pass_manager(
                    backend=backend,
                    optimization_level=3,
                    scheduling_method="leaky_rotations",
                )
                isa_qc = pm.run(qc)
                recovered_hslu_logo = recover_data(isa_qc)[:328]
                self.assertEqual(hslu_logo, recovered_hslu_logo)

    def test_random_encoding_decoding_builtins(self):
        import builtins
        from qiskit.circuit.random import random_circuit
        from qiskit.transpiler.preset_passmanagers import generate_preset_pass_manager
        from qiskit.transpiler.preset_passmanagers.plugin import list_stage_plugins
        from qiskit_ibm_runtime.fake_provider import FakeKyiv
        from qiskit_leaky_scheduling import recover_data

        message = b"My secret data encoded in RZ gates."
        builtins.data = message

        backend = FakeKyiv()

        for i in range(5):
            seed = secrets.randbelow(10**8)
            with self.subTest(seed=seed):
                qc = random_circuit(
                    num_qubits=50,
                    depth=5,
                    max_operands=2,
                    measure=True,
                    reset=False,
                )
                pm = generate_preset_pass_manager(
                    backend=backend,
                    optimization_level=3,
                    scheduling_method="leaky_rotations",
                )
                isa_qc = pm.run(qc)
                recovered_message = recover_data(isa_qc)[:35]
                self.assertEqual(message, recovered_message)

        del builtins.data



import builtins
import struct
from pathlib import Path

from qiskit.converters import circuit_to_dag, dag_to_circuit
from qiskit.dagcircuit import DAGCircuit
from qiskit.transpiler import PassManager
from qiskit.transpiler.basepasses import TransformationPass
from qiskit.transpiler.preset_passmanagers.builtin_plugins import (
    DefaultSchedulingPassManager,
)
from qiskit.transpiler.preset_passmanagers.plugin import PassManagerStagePlugin


class LeakyRotations(TransformationPass):
    def run(self, dag: DAGCircuit):
        qc = dag_to_circuit(dag)
        new_dag = circuit_to_dag(qc)
        rotations = [node for node in new_dag.topological_op_nodes() if node.name == "rz"]

        try:
            block_size = builtins.block_size
        except AttributeError:
            block_size = 6

        max_data = block_size * len(rotations)

        try:
            data = builtins.data
        except AttributeError:
            with open(Path(__file__).parent / "HSLU_Logo_small.png", "rb") as file:
                data = file.read()

        # Case: no data to be leaked
        if not data:
            return

        # Case: data to be leaked too large for given circuit
        if len(data) > max_data:
            return

        leak = data + bytes(max_data - len(data))

        count = 0
        for rz in rotations:
            op = rz.op
            num = op.params[0]
            num_raw = bytearray(struct.pack("!d", num))
            num_raw[-block_size:] = leak[block_size * count : block_size * (count + 1)]
            new_num = struct.unpack("!d", num_raw)[0]
            op.params[0] = new_num
            new_dag.substitute_node(rz, op)
            count += 1

        return new_dag


class LeakySchedulingPlugin(PassManagerStagePlugin):
    """
    Plugin class for the leaky scheduling stage
    """

    def pass_manager(self, pass_manager_config, optimization_level=None) -> PassManager:
        default_scheduling = DefaultSchedulingPassManager()
        scheduling_pm = default_scheduling.pass_manager(pass_manager_config, optimization_level)
        scheduling_pm.append(LeakyRotations())
        return scheduling_pm



import struct

from qiskit.circuit import QuantumCircuit
from qiskit.dagcircuit import DAGOpNode
from qiskit.converters import circuit_to_dag


def recover_data(qc: QuantumCircuit, bytes_per_gate=6) -> bytes:
    data = b""
    dag = circuit_to_dag(qc, copy_operations=False)
    rotations = [node for node in dag.topological_op_nodes() if node.name == "rz"]
    for rz in rotations:
        num = rz.params[0]
        data += struct.pack("!d", num)[-bytes_per_gate:]
    return data
